#include <config.h>
#include <sys/types.h>
#include <stdio.h>
#include <setjmp.h>


#ifdef HAVE_CMIGEMO
#include "lisp.h"
#include "charset.h"
#include "coding.h"

#ifdef WINDOWSNT
#include <windows.h>
#define MIGEMO_DLL "migemo.dll"
#else
#define INVALID_HANDLE_VALUE NULL
#define MIGEMO_DLL "libmigemo.so"
#include <dlfcn.h>
#endif

#include "migemo.h"

static migemo* (*migemo_open_)(const char*);
static void (*migemo_close_)(migemo*);
static unsigned char* (*migemo_query_)(migemo*, unsigned char*);
static void (*migemo_release_)(migemo*, unsigned char*);

static int (*migemo_set_operator_)(migemo*, int, unsigned char*);
static int (*migemo_is_enable_)(migemo*);
static int (*migemo_load_)(migemo*, int, char *);
/*
static void (*migemo_setproc_char2int_)(migemo*, MIGEMO_PROC_CHAR2INT);
static void (*migemo_setproc_int2char_)(migemo*, MIGEMO_PROC_INT2CHAR);
*/

static migemo *pmigemo;
#ifdef WINDOWSNT
static HMODULE hMigemo;
#else
static void *hMigemo;
#endif

static int eucjp_int2char(unsigned int in, unsigned char* out)
{
    /* CP932と内容は同じだが将来JISX0213に対応させるために分離しておく */
    int len = 0;

    if (!out)
	return len;

    switch (in)  {
    case '.': case '*': case '+': case '?':
    case '[': case ']': case '^': case '$':
    case '\\':
	out[len++] = '\\';
	/* fall through */
    default:
	if (in >= 0x100)
	    out[len++] = (unsigned char)((in >> 8) & 0xFF);
	out[len++] = (unsigned char)(in & 0xFF);
	break;
    }
    return len;
}

DEFUN ("cmigemo-open", Fcmigemo_open, Scmigemo_open, 1, 1, 0,
       "open cmigemo with DICTIONARY.")
     (dict)
     Lisp_Object dict;
{
  static int coding_system_initialized = 0;

  if (pmigemo)
    error ("cmigemo already opened");

  CHECK_STRING(dict);

#ifdef WINDOWSNT
  hMigemo = LoadLibrary(MIGEMO_DLL);
#else
  hMigemo = dlopen(MIGEMO_DLL, RTLD_NOW);
#endif
  if (hMigemo == INVALID_HANDLE_VALUE)
    error (MIGEMO_DLL ": can't load");

#ifdef WINDOWSNT
#define GETPROC(n, t) do { \
	migemo_ ## n ## _ = (t) GetProcAddress(hMigemo, "migemo_" # n); \
} while(0)
#else
#define GETPROC(n, t) do { \
	migemo_ ## n ## _ = (t) dlsym(hMigemo, "migemo_" # n); \
} while(0)
#endif

  GETPROC(open, migemo* (*)(const char*));
  GETPROC(close, void (*)(migemo*));
  GETPROC(query, unsigned char* (*)(migemo*, unsigned char*));
  GETPROC(release, void (*)(migemo*, unsigned char*));
  GETPROC(set_operator, int (*)(migemo*, int, unsigned char*));
  GETPROC(is_enable, int (*)(migemo*));
  GETPROC(load, int (*)(migemo*, int, char *));
/*
  GETPROC(setproc_char2int, void (*)(migemo*, MIGEMO_PROC_CHAR2INT));
  GETPROC(setproc_int2char, void (*)(migemo*, MIGEMO_PROC_INT2CHAR));
*/
  if (!(pmigemo = migemo_open_(XSTRING(dict)->data)) ||
      !migemo_is_enable_(pmigemo)) {
    pmigemo = 0;
    error("could not open cmigemo");
  }
  migemo_set_operator_(pmigemo, MIGEMO_OPINDEX_OR, "\\|");
  migemo_set_operator_(pmigemo, MIGEMO_OPINDEX_NEST_IN, "\\(");
  migemo_set_operator_(pmigemo, MIGEMO_OPINDEX_NEST_OUT, "\\)");
  migemo_set_operator_(pmigemo, MIGEMO_OPINDEX_NEWLINE, "\\s-*");
/*
  migemo_setproc_int2char_(pmigemo, eucjp_int2char);
*/
  return Qt;
}

DEFUN ("cmigemo-close", Fcmigemo_close, Scmigemo_close, 0, 0, 0,
       "close cmigemo.")
     ()
{
  if (!pmigemo)
    error ("cmigemo was not opened");
  migemo_close_(pmigemo);
  pmigemo = 0;
#ifdef WINDOWSNT
  FreeLibrary(hMigemo);
#else
  dlclose(hMigemo);
#endif
  hMigemo = 0;
  return Qt;
}

DEFUN ("cmigemo-query", Fcmigemo_query, Scmigemo_query, 1, 1, 0,
       "query cmigemo about WORD.")
     (word)
     Lisp_Object word;
{
  char *ans;
  Lisp_Object ret;

  if (!pmigemo)
    error ("cmigemo was not opened");

  CHECK_STRING(word);

  ans = migemo_query_(pmigemo, XSTRING(word)->data);
  if (!ans)
    ret = Qnil;
  else
    ret = code_convert_string(build_string(ans),intern("euc-japan"), 
			      Qnil, 0, 1, 1);
  migemo_release_(pmigemo, ans);
  return ret;
}

DEFUN ("cmigemo-load", Fcmigemo_load, Scmigemo_load, 1, 1, 0,
       "load a sub DICTIONARY.")
     (dict)
     Lisp_Object dict;
{
  if (!pmigemo)
    error ("cmigemo was not opened");

  CHECK_STRING(dict);

  if (migemo_load_(pmigemo,
		   MIGEMO_DICTID_MIGEMO,
		   XSTRING(dict)->data) == MIGEMO_DICTID_INVALID)
    error("migemo_load invalid dict %s", dict);

/*
  migemo_setproc_int2char_(pmigemo, eucjp_int2char);
*/
  return Qt;
}

void
syms_of_cmigemo ()
{
  defsubr(&Scmigemo_open);
  defsubr(&Scmigemo_close);
  defsubr(&Scmigemo_query);
  defsubr(&Scmigemo_load);
}
#endif /* HAVE_CMIGEMO */
